/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

// draw.c -- this is the only file outside the refresh that touches the
// vid buffer

#include "quakedef.h"

byte		*draw_chars;				// 8*8 graphic characters
qpic_t		*draw_disc;
qpic_t		*draw_backtile;
qpic_t		*conback;

int			translate_texture;
gltexture_t	*char_texture;

gltexture_t *GL_LoadPicTexture (char *identifier, qpic_t *pic);
void GL_TextureInit (void);


/*
=============================================================================

  scrap allocation

  Allocate all the little status bar obejcts into a single texture
  to crutch up stupid hardware / drivers

=============================================================================
*/

// increased MAX_SCRAPS from 2 to 4 to account for border issues
#define	MAX_SCRAPS		4

// changed these defines so they don't conflict with the lightmaps!!!
#define	SCRAP_BLOCK_WIDTH	256
#define	SCRAP_BLOCK_HEIGHT	256

int			scrap_allocated[MAX_SCRAPS][SCRAP_BLOCK_WIDTH];
byte		scrap_texels[MAX_SCRAPS][SCRAP_BLOCK_WIDTH * SCRAP_BLOCK_HEIGHT * 4];
qboolean	scrap_dirty[MAX_SCRAPS];
int			scrap_texnum;
gltexture_t scrapGlts[MAX_SCRAPS];


// returns a texture number and the position inside it
int Scrap_AllocBlock (int w, int h, int *x, int *y)
{
	int		i, j;
	int		best, best2;
	int		bestx;
	int		texnum;

	for (texnum = 0; texnum < MAX_SCRAPS; texnum++)
	{
		best = SCRAP_BLOCK_HEIGHT;

		for (i = 0; i < SCRAP_BLOCK_WIDTH - w; i++)
		{
			best2 = 0;

			for (j = 0; j < w; j++)
			{
				if (scrap_allocated[texnum][i+j] >= best)
					break;
				if (scrap_allocated[texnum][i+j] > best2)
					best2 = scrap_allocated[texnum][i+j];
			}
			if (j == w)
			{	// this is a valid spot
				*x = i;
				*y = best = best2;
			}
		}

		if (best + h > SCRAP_BLOCK_HEIGHT)
			continue;

		for (i=0 ; i<w ; i++)
			scrap_allocated[texnum][*x + i] = best + h;

		return texnum;
	}

	// game will crash here, so...
	Sys_Error ("Scrap_AllocBlock: full");

	// ...shut up the compiler
	return 0;
}

int	scrap_uploads;

void Scrap_Upload (void)
{
	int		texnum;

	for (texnum = 0; texnum < MAX_SCRAPS; texnum++)
	{
		// only upload dirty scraps
		if (scrap_dirty[texnum])
		{
			// Con_Printf ("Scrap Upload\n");

			GL_UploadTexture (&scrapGlts[texnum], scrap_texels[texnum], SCRAP_BLOCK_WIDTH, SCRAP_BLOCK_HEIGHT, 8, TEX_2D);
			scrap_dirty[texnum] = false;

			scrap_uploads++;

			// output to TGA for testing
			// WriteTGA (va ("scrap_%01i.tga", texnum), SCRAP_BLOCK_WIDTH, SCRAP_BLOCK_HEIGHT, scrap_texels[texnum], 8);
		}
	}
}

//=============================================================================
/* Support Routines */

typedef struct cachepic_s
{
	char		name[MAX_QPATH];
	qpic_t		pic;
	byte		padding[32];	// for appended glpic
} cachepic_t;

#define	MAX_CACHED_PICS		128
cachepic_t	menu_cachepics[MAX_CACHED_PICS];
int			menu_numcachepics;

byte		menuplyr_pixels[4096];

int		pic_texels;
int		pic_count;

qpic_t *Draw_PicFromWad (char *name)
{
	qpic_t	*p;
	glpic_t	*gl;

	p = W_GetLumpName (name);
	gl = (glpic_t *) p->data;

	// load little ones into the scrap
	if (p->width < 64 && p->height < 64)
	{
		int		x, y;
		int		i, j, k;
		int		texnum;

		// give a border of 8 around the tex to avoid blend fringes
		texnum = Scrap_AllocBlock (p->width + 8, p->height + 8, &x, &y);
		scrap_dirty[texnum] = true;

		// set the scrap block to all alpha
		for (i = 0; i < p->height + 8; i++)
			for (j = 0; j < p->width + 8; j++)
				scrap_texels[texnum][(y + i) * SCRAP_BLOCK_WIDTH + x + j] = 255;

		// copy in the texture with a border of 4 texels
		for (i = 4, k = 0; i < p->height + 4; i++)
			for (j = 4; j < p->width + 4; j++, k++)
				scrap_texels[texnum][(y + i) * SCRAP_BLOCK_WIDTH + x + j] = p->data[k];

		// fill in the glt and it's properties
		// these will be hit every tim the current scrap texture is loaded, but what the fuck?
		gl->glt = &scrapGlts[texnum];
		gl->glt->texnum = scrap_texnum + texnum;
		gl->glt->flags = (TEX_2D | TEX_SCRAP);
		gl->glt->width = SCRAP_BLOCK_WIDTH;
		gl->glt->height = SCRAP_BLOCK_HEIGHT;
		gl->glt->offset = 0;

		// these aren't really needed but we'll do them anyway...
		// put invalid data into the names
		sprintf (gl->glt->sourcefile, ">**scrap_%i**<", texnum);
		sprintf (gl->glt->identifier, ">**scrap_%i**<", texnum);
		MD5_Checksum (scrap_texels[texnum], SCRAP_BLOCK_WIDTH * SCRAP_BLOCK_HEIGHT, gl->glt->md5digest);

		// set the texcoords
		gl->sl = (float) (x + 4) / (float) SCRAP_BLOCK_WIDTH;
		gl->sh = (float) (x + 4 + p->width) / (float) SCRAP_BLOCK_WIDTH;
		gl->tl = (float) (y + 4) / (float) SCRAP_BLOCK_WIDTH;
		gl->th = (float) (y + 4 + p->height) / (float) SCRAP_BLOCK_WIDTH;

		pic_count++;
		pic_texels += p->width * p->height;
	}
	else
	{
		gl->glt = GL_LoadPicTexture (name, p);
		gl->sl = 0;
		gl->sh = 1;
		gl->tl = 0;
		gl->th = 1;
	}

	return p;
}


/*
================
Draw_CachePic
================
*/
qpic_t	*Draw_CachePic (char *path)
{
	cachepic_t	*pic;
	int			i;
	qpic_t		*dat;
	glpic_t		*gl;

	for (pic=menu_cachepics, i=0 ; i<menu_numcachepics ; pic++, i++)
		if (!strcmp (path, pic->name))
			return &pic->pic;

	if (menu_numcachepics == MAX_CACHED_PICS)
		Sys_Error ("menu_numcachepics == MAX_CACHED_PICS");
	menu_numcachepics++;
	strcpy (pic->name, path);

	// load the pic from disk
	dat = (qpic_t *)COM_LoadTempFile (path);	
	if (!dat)
		Sys_Error ("Draw_CachePic: failed to load %s", path);

	SwapPic (dat);

	// HACK HACK HACK --- we need to keep the bytes for
	// the translatable player picture just for the menu
	// configuration dialog
	if (!strcmp (path, "gfx/menuplyr.lmp"))
		Q_memcpy (menuplyr_pixels, dat->data, dat->width*dat->height);

	pic->pic.width = dat->width;
	pic->pic.height = dat->height;

	gl = (glpic_t *) pic->pic.data;

	// load little ones into the scrap
	if (pic->pic.width < 64 && pic->pic.height < 64)
	{
		int		x, y;
		int		i, j, k;
		int		texnum;

		// give a border of 8 around the tex to avoid blend fringes
		texnum = Scrap_AllocBlock (pic->pic.width + 8, pic->pic.height + 8, &x, &y);
		scrap_dirty[texnum] = true;

		// set the scrap block to all alpha
		for (i = 0; i < pic->pic.height + 8; i++)
			for (j = 0; j < pic->pic.width + 8; j++)
				scrap_texels[texnum][(y + i) * SCRAP_BLOCK_WIDTH + x + j] = 255;

		// filthy hack - copy the full thing 4 times with an offset of 2 to avoid texture border crapness
		// it's still not perfect, but a hell of a lot better than it was...!
		for (i = 2, k = 0; i < pic->pic.height + 2; i++)
			for (j = 2; j < pic->pic.width + 2; j++, k++)
				scrap_texels[texnum][(y + i) * SCRAP_BLOCK_WIDTH + x + j] = dat->data[k];

		for (i = 2, k = 0; i < pic->pic.height + 2; i++)
			for (j = 6; j < pic->pic.width + 6; j++, k++)
				scrap_texels[texnum][(y + i) * SCRAP_BLOCK_WIDTH + x + j] = dat->data[k];

		for (i = 6, k = 0; i < pic->pic.height + 6; i++)
			for (j = 6; j < pic->pic.width + 6; j++, k++)
				scrap_texels[texnum][(y + i) * SCRAP_BLOCK_WIDTH + x + j] = dat->data[k];

		for (i = 6, k = 0; i < pic->pic.height + 6; i++)
			for (j = 2; j < pic->pic.width + 2; j++, k++)
				scrap_texels[texnum][(y + i) * SCRAP_BLOCK_WIDTH + x + j] = dat->data[k];

		// now copy again with an offset of 1 so that the very outside texels are replicated twice
		// i could have done this properly and only copied the required texels but this way is easier, and
		// since it doesn't happen at runtime functional is all that matters
		for (i = 3, k = 0; i < pic->pic.height + 3; i++)
			for (j = 3; j < pic->pic.width + 3; j++, k++)
				scrap_texels[texnum][(y + i) * SCRAP_BLOCK_WIDTH + x + j] = dat->data[k];

		for (i = 3, k = 0; i < pic->pic.height + 3; i++)
			for (j = 5; j < pic->pic.width + 5; j++, k++)
				scrap_texels[texnum][(y + i) * SCRAP_BLOCK_WIDTH + x + j] = dat->data[k];

		for (i = 5, k = 0; i < pic->pic.height + 5; i++)
			for (j = 5; j < pic->pic.width + 5; j++, k++)
				scrap_texels[texnum][(y + i) * SCRAP_BLOCK_WIDTH + x + j] = dat->data[k];

		for (i = 5, k = 0; i < pic->pic.height + 5; i++)
			for (j = 3; j < pic->pic.width + 3; j++, k++)
				scrap_texels[texnum][(y + i) * SCRAP_BLOCK_WIDTH + x + j] = dat->data[k];

		// now copy in the texture with a border of 4 texels
		for (i = 4, k = 0; i < pic->pic.height + 4; i++)
			for (j = 4; j < pic->pic.width + 4; j++, k++)
				scrap_texels[texnum][(y + i) * SCRAP_BLOCK_WIDTH + x + j] = dat->data[k];

		// fill in the glt and it's properties
		// these will be hit every tim the current scrap texture is loaded, but what the fuck?
		gl->glt = &scrapGlts[texnum];
		gl->glt->texnum = scrap_texnum + texnum;
		gl->glt->flags = (TEX_2D | TEX_SCRAP);
		gl->glt->width = SCRAP_BLOCK_WIDTH;
		gl->glt->height = SCRAP_BLOCK_HEIGHT;
		gl->glt->offset = 0;

		// these aren't really needed but we'll do them anyway...
		// put invalid data into the names
		sprintf (gl->glt->sourcefile, ">**scrap_%i**<", texnum);
		sprintf (gl->glt->identifier, ">**scrap_%i**<", texnum);
		MD5_Checksum (scrap_texels[texnum], SCRAP_BLOCK_WIDTH * SCRAP_BLOCK_HEIGHT, gl->glt->md5digest);

		// set the texcoords
		gl->sl = (float) (x + 4) / (float) SCRAP_BLOCK_WIDTH;
		gl->sh = (float) (x + 4 + pic->pic.width) / (float) SCRAP_BLOCK_WIDTH;
		gl->tl = (float) (y + 4) / (float) SCRAP_BLOCK_WIDTH;
		gl->th = (float) (y + 4 + pic->pic.height) / (float) SCRAP_BLOCK_WIDTH;

		pic_count++;
		pic_texels += pic->pic.width * pic->pic.height;
	}
	else
	{
		gl->glt = GL_LoadPicTexture (path, dat);
		gl->sl = 0;
		gl->sh = 1;
		gl->tl = 0;
		gl->th = 1;
	}

	return &pic->pic;
}


void WriteTGA (char *name, int w, int h, byte *data, int bpp)
{
	FILE *f;
	int i;
	byte buffer[18];

	if (!(bpp == 8 || bpp == 32)) return;
	f = fopen (name, "wb");
	if (!f) return;

	memset (buffer, 0, 18);
	buffer[2] = 2;
	buffer[12] = w & 255;
	buffer[13] = w >> 8;
	buffer[14] = h & 255;
	buffer[15] = h >> 8;
	buffer[16] = 32;
	buffer[17] = 32;

	fwrite (buffer, 18, 1, f);

	if (bpp == 32)
		fwrite (data, w * h * 4, 1, f);
	else
	{
		for (i = 0; i < w * h; i++)
		{
			// bgra palette for tga
			fwrite (&d_8to24tableBGRA[data[i]], 4, 1, f);
		}
	}

	fclose (f);
}

void InitBaseTextures (void)
{
	int i;
	int j;
	int c;
	byte *newCharset;

	// load the charset manually because we need to hack the charset alpha colour and character
	// spacing because one of id's artists badly needs a good slap around the head...
	draw_chars = W_GetLumpName ("conchars");

	// set the proper translucent colour
	for (i = 0; i < 16384; i++)
	{
		if (draw_chars[i] == 0)
		{
			draw_chars[i] = 255;
		}
	}

	// allocate memory for a new charset
	newCharset = (byte *) malloc (256 * 256);

	// clear to all alpha
	memset (newCharset, 255, 256 * 256);

	// copy them in with better spacing
	for (i = 0, j = 0; i < 128 * 128; i += 8, j += 16)
	{
		// jump down a half row
		if (i && !(i % 1024)) j += 2048;

		for (c = 0; c < 8; c++)
		{
			newCharset[j + c] = draw_chars[i + c];
		}
	}

	// now this is evil.
	// some of the chars are meant as consecutive dividers and the like, so we don;t want them spaced.
	// because we're spacing by the width of a char, we just copy the relevant pixels in
	for (i = 16; i < 24; i++)
	{
		for (j = 224; j < 232; j++)
		{
			newCharset[i * 256 + j - 8] = newCharset[i * 256 + j];
			newCharset[i * 256 + j + 8] = newCharset[i * 256 + j];
		}
	}

	for (i = 144; i < 152; i++)
	{
		for (j = 224; j < 232; j++)
		{
			newCharset[i * 256 + j - 8] = newCharset[i * 256 + j];
			newCharset[i * 256 + j + 8] = newCharset[i * 256 + j];
		}
	}

	for (i = 128; i < 136; i++)
	{
		for (j = 16; j < 24; j++)
		{
			newCharset[i * 256 + j - 8] = newCharset[i * 256 + j];
			newCharset[i * 256 + j + 8] = newCharset[i * 256 + j];
		}
	}

	for (i = 128; i < 136; i++)
	{
		newCharset[i * 256 + 47] = newCharset[i * 256 + 23];
		newCharset[i * 256 + 56] = newCharset[i * 256 + 16];
	}

	// now turn them into textures
	char_texture = GL_LoadTexture ("charset", 256, 256, newCharset, TEX_2D);

	// free memory
	free (newCharset);

	// load console
	conback = Draw_CachePic ("gfx/conback.lmp");

	// hack the properties so it draws correctly
	conback->height = vid.height;
	conback->width = vid.width;

	// get the other pics we need
	draw_disc = Draw_PicFromWad ("disc");
	draw_backtile = Draw_PicFromWad ("backtile");
}


/*
===============
Draw_Init
===============
*/
void Draw_Init (void)
{
	GL_TextureInit ();

	// save a texture slot for translated picture
	translate_texture = texture_extension_number++;

	// save slots for scraps
	scrap_texnum = texture_extension_number;
	texture_extension_number += MAX_SCRAPS;

	InitBaseTextures ();
}



/*
================
Draw_Character

Draws one 8*8 graphics character with 0 being transparent.
It can be clipped to the top of the screen to allow the console to be
smoothly scrolled off.
================
*/
void Draw_Character (int x, int y, int num)
{
	byte			*dest;
	byte			*source;
	unsigned short	*pusdest;
	int				drawline;	
	int				row, col;
	float			frow, fcol, size;

	// space
	if (num == 32) return;

	num &= 255;

	// totally off screen
	if (y <= -8) return;

	row = num >> 4;
	col = num & 15;

	frow = row * 0.0625;
	fcol = col * 0.0625;
	size = 0.03125;

	glBindTexture (GL_TEXTURE_2D, char_texture->texnum);

	glBegin (GL_QUADS);
	glTexCoord2f (fcol, frow);
	glVertex2f (x, y);
	glTexCoord2f (fcol + size, frow);
	glVertex2f (x + 8, y);
	glTexCoord2f (fcol + size, frow + size);
	glVertex2f (x + 8, y + 8);
	glTexCoord2f (fcol, frow + size);
	glVertex2f (x, y + 8);
	glEnd ();
}

void Draw_StringCharacter (int x, int y, int num)
{
	byte			*dest;
	byte			*source;
	unsigned short	*pusdest;
	int				drawline;	
	int				row, col;
	float			frow, fcol, size;

	// space
	if (num == 32) return;

	// clamp to scale
	num &= 255;

	// totally off screen
	if (y <= -8) return;

	row = num >> 4;
	col = num & 15;

	frow = row * 0.0625;
	fcol = col * 0.0625;
	size = 0.03125;

	glTexCoord2f (fcol, frow);
	glVertex2f (x, y);
	glTexCoord2f (fcol + size, frow);
	glVertex2f (x + 8, y);
	glTexCoord2f (fcol + size, frow + size);
	glVertex2f (x + 8, y + 8);
	glTexCoord2f (fcol, frow + size);
	glVertex2f (x, y + 8);
}

/*
================
Draw_String
================
*/
void Draw_String (int x, int y, char *str)
{
	glBindTexture (GL_TEXTURE_2D, char_texture->texnum);

	glBegin (GL_QUADS);

	while (*str)
	{
		Draw_StringCharacter (x, y, *str);
		str++;
		x += 8;
	}

	glEnd ();
}

/*
================
Draw_DebugChar

Draws a single character directly to the upper right corner of the screen.
This is for debugging lockups by drawing different chars in different parts
of the code.
================
*/
void Draw_DebugChar (char num)
{
}

/*
=============
Draw_AlphaPic
=============
*/
void Draw_AlphaPic (int x, int y, qpic_t *pic, float alpha)
{
	byte			*dest, *source;
	unsigned short	*pusdest;
	int				v, u;
	glpic_t			*gl;

	Scrap_Upload ();

	gl = (glpic_t *)pic->data;

	glColor4f (1,1,1,alpha);
	glBindTexture (GL_TEXTURE_2D, gl->glt->texnum);
	glBegin (GL_QUADS);
	glTexCoord2f (gl->sl, gl->tl);
	glVertex2f (x, y);
	glTexCoord2f (gl->sh, gl->tl);
	glVertex2f (x+pic->width, y);
	glTexCoord2f (gl->sh, gl->th);
	glVertex2f (x+pic->width, y+pic->height);
	glTexCoord2f (gl->sl, gl->th);
	glVertex2f (x, y+pic->height);
	glEnd ();
	glColor4f (1,1,1,1);
}


/*
=============
Draw_Pic
=============
*/
void Draw_Pic (int x, int y, qpic_t *pic)
{
	glpic_t			*gl;

	Scrap_Upload ();

	gl = (glpic_t *) pic->data;

	glColor4f (1,1,1,1);

	glBindTexture (GL_TEXTURE_2D, gl->glt->texnum);

	glBegin (GL_QUADS);

	glTexCoord2f (gl->sl, gl->tl);
	glVertex2f (x, y);

	glTexCoord2f (gl->sh, gl->tl);
	glVertex2f (x+pic->width, y);

	glTexCoord2f (gl->sh, gl->th);
	glVertex2f (x+pic->width, y+pic->height);

	glTexCoord2f (gl->sl, gl->th);
	glVertex2f (x, y+pic->height);

	glEnd ();
}


void Draw_ScaledPic (int x, int y, qpic_t *pic, float scale)
{
	glpic_t			*gl;

	Scrap_Upload ();

	gl = (glpic_t *) pic->data;

	glColor4f (1, 1, 1, 1);

	glBindTexture (GL_TEXTURE_2D, gl->glt->texnum);

	glBegin (GL_QUADS);

	glTexCoord2f (gl->sl, gl->tl);
	glVertex2f (x, y);

	glTexCoord2f (gl->sh, gl->tl);
	glVertex2f (x + ((float) pic->width * scale), y);

	glTexCoord2f (gl->sh, gl->th);
	glVertex2f (x + ((float) pic->width * scale), y + ((float) pic->height * scale));

	glTexCoord2f (gl->sl, gl->th);
	glVertex2f (x, y + ((float) pic->height * scale));

	glEnd ();
}


/*
=============
Draw_TransPic

Fixme: remove this altogether.
=============
*/
void Draw_TransPic (int x, int y, qpic_t *pic)
{
	Draw_Pic (x, y, pic);
}


/*
=============
Draw_TransPicTranslate

Only used for the player color selection menu
=============
*/
void Draw_TransPicTranslate (int x, int y, qpic_t *pic, byte *translation)
{
	int				v, u, c;
	unsigned		trans[64*64], *dest;
	byte			*src;
	int				p;

	glBindTexture (GL_TEXTURE_2D, translate_texture);

	c = pic->width * pic->height;

	dest = trans;
	for (v=0 ; v<64 ; v++, dest += 64)
	{
		src = &menuplyr_pixels[ ((v*pic->height)>>6) *pic->width];
		for (u=0 ; u<64 ; u++)
		{
			p = src[(u*pic->width)>>6];
			if (p == 255)
				dest[u] = p;
			else
				dest[u] =  d_8to24tableBGRA[translation[p]];
		}
	}

	glTexImage2D (GL_TEXTURE_2D, 0, GL_RGBA, 64, 64, 0, GL_BGRA, GL_UNSIGNED_BYTE, trans);

	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	glColor3f (1,1,1);
	glBegin (GL_QUADS);
	glTexCoord2f (0, 0);
	glVertex2f (x, y);
	glTexCoord2f (1, 0);
	glVertex2f (x+pic->width, y);
	glTexCoord2f (1, 1);
	glVertex2f (x+pic->width, y+pic->height);
	glTexCoord2f (0, 1);
	glVertex2f (x, y+pic->height);
	glEnd ();
}


/*
================
Draw_ConsoleBackground

================
*/
void Draw_ConsoleBackground (int lines)
{
	int y = (vid.height2D * 3) >> 2;

	if (lines > y)
		Draw_Pic(0, lines - vid.height2D, conback);
	else
		Draw_AlphaPic (0, lines - vid.height2D, conback, (float)(1.2 * lines)/y);
}


/*
=============
Draw_Fill

Fills a box of pixels with a single color
=============
*/
void Draw_Fill (int x, int y, int w, int h, int c)
{
	glDisable (GL_TEXTURE_2D);
	glColor3f (host_basepal[c*3]/255.0,
		host_basepal[c*3+1]/255.0,
		host_basepal[c*3+2]/255.0);

	glBegin (GL_QUADS);

	glVertex2f (x,y);
	glVertex2f (x+w, y);
	glVertex2f (x+w, y+h);
	glVertex2f (x, y+h);

	glEnd ();
	glColor3f (1,1,1);
	glEnable (GL_TEXTURE_2D);
}
//=============================================================================

/*
================
Draw_FadeScreen

================
*/
void Draw_FadeScreen (void)
{
	glDisable (GL_TEXTURE_2D);
	glColor4f (0, 0, 0, 0.8);
	glBegin (GL_QUADS);

	glVertex2f (0,0);
	glVertex2f (vid.width2D, 0);
	glVertex2f (vid.width2D, vid.height2D);
	glVertex2f (0, vid.height2D);

	glEnd ();
	glColor4f (1,1,1,1);
	glEnable (GL_TEXTURE_2D);
}

//=============================================================================

/*
================
Draw_BeginDisc

Draws the little blue disc in the corner of the screen.
Call before beginning any disc IO.
================
*/
void Draw_BeginDisc (void)
{
	// removed
}


/*
================
Draw_EndDisc

Erases the disc icon.
Call after completing any disc IO
================
*/
void Draw_EndDisc (void)
{
}


void Draw_CalcRefdef (void)
{
	float Width2D, Height2D;
	float aspect = (float) glwidth / (float) glheight;

	vid.recalc_2drefdef = false;

	if (aspect > (4.0f / 3.0f))
	{
		// widescreen.  keep height, adjust width
		Height2D = 480;
		Width2D = aspect * Height2D;
	}
	else
	{
		// keep width, adjust height
		Width2D = 640;
		Height2D = Width2D / aspect;
	}

	// adjust dimensions
	vid.width2D = Width2D;
	vid.height2D = Height2D;

	// adjust size of conback picture
	conback->width = vid.width2D;
	conback->height = vid.height2D;

	// force a resize of the console buffer
	// Con_CheckResize (true);
}


/*
================
GL_Set2D

Setup as if the screen was 320*200
================
*/
void GL_Set2D (void)
{
	if (vid.recalc_2drefdef) Draw_CalcRefdef ();

	glViewport (0, 0, glwidth, glheight);

	glMatrixMode(GL_PROJECTION);
    glLoadIdentity ();
	glOrtho (0, vid.width2D, vid.height2D, 0, -99999, 99999);

	glMatrixMode(GL_MODELVIEW);
    glLoadIdentity ();

	glDisable (GL_DEPTH_TEST);
	glDisable (GL_CULL_FACE);

	// do with blending instead of alpha testing to avoid border fringes
	glEnable (GL_BLEND);
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	glColor4f (1, 1, 1, 1);
}

